// movie.cpp
//
// Implementation for CMovie
//

#include "stdafx.h"
#include "Stage.h"
#include "movie.h"

///////////////////////////////////////////////////////////////////////////////
// CMovieFrame

CMovieFrame::CMovieFrame(HBITMAP hBmp, BITMAPINFO* pBMI, void* pBits)
{
	m_hBmp = hBmp;
	m_pBMI = pBMI;
	m_pBits = pBits;
}

CMovieFrame::~CMovieFrame()
{
	if (m_hBmp) {
		::DeleteObject(m_hBmp);
		ASSERT(m_pBMI);
		delete m_pBMI;
	}
}

///////////////////////////////////////////////////////////////////////////////
// CMovieTake

CMovieTake::CMovieTake()
{
	m_pPalette = NULL;
	m_iHeight = 0;
	m_iWidth = 0;
}

CMovieTake::~CMovieTake()
{
	DeleteAll();
	if (m_pPalette) delete m_pPalette;
}

void CMovieTake::DeleteAll()
{
	while (!IsEmpty()) {
		CMovieFrame* pFrame = RemoveHead();
		ASSERT(pFrame);
		delete pFrame;
	}
}

void CMovieTake::SetPalette(CPalette* pPalette)
{
	if (m_pPalette) delete m_pPalette;
	m_pPalette = pPalette;
}

void CMovieTake::Optimize()
{
	// make sure we have a palette to map to
	if (!m_pPalette) return;

	// Get the first frame
	CMovieFrame* pFrame = GetHead();
	ASSERT(pFrame);

	// make sure it's 8 bpp
	if (pFrame->m_pBMI->bmiHeader.biBitCount != 8) return;
	
	// Make an identity palette
    int i, iSysColors, iPalEntries;
    HPALETTE hpalOld;

    // Get a screen DC to work with
    HWND hwndActive = ::GetActiveWindow();
    HDC hdcScreen = ::GetDC(hwndActive);
    ASSERT(hdcScreen);

    // Make sure we are on a palettized device
    if (!::GetDeviceCaps(hdcScreen, RASTERCAPS) & RC_PALETTE) {
        return;
    }

    // Get the number of system colors and the number of palette entries
    // Note that on a palletized device the number of colors is the
    // number of guaranteed colors.  I.e. the number of reserved system colors
    iSysColors = ::GetDeviceCaps(hdcScreen, NUMCOLORS);
    iPalEntries = ::GetDeviceCaps(hdcScreen, SIZEPALETTE);

    // if there are more than 256 colors we are wasting our time
    if (iSysColors > 256) return;

    // Now we force the palette manager to reset its tables so that
    // the next palette to be realized will get its colors in the order they are 
    // in the logical palette. This is done by changing the number of
    // reserved colors.
    ::SetSystemPaletteUse(hdcScreen, SYSPAL_NOSTATIC);
    ::SetSystemPaletteUse(hdcScreen, SYSPAL_STATIC);

    // Select our palette into the screen DC and realize it so that
    // its colors will be entered into the free slots in the physical palette
    hpalOld = ::SelectPalette(hdcScreen,
                              (HPALETTE) (*m_pPalette),
                              FALSE);
    ::RealizePalette(hdcScreen);
    // Now replace the old palette (but don't realize it)
    ::SelectPalette(hdcScreen, hpalOld, FALSE);

    // The physical palette now has our colors set in place and its own
    // reserved colors at either end.  We can grab the lot now
    PALETTEENTRY pe[256];
    ::GetSystemPaletteEntries(hdcScreen, 
                            0,
                            iPalEntries,
                            pe);

    // Set the PC_NOCOLLAPSE flag for each of our colors so that GDI
    // won't merge them together.  Be careful not to set PC_NOCOLLAPSE for the 
    // sys color entries or we'll get multpile copies of these colors in
    // the palette when we realize it.
    for (i = 0; i < iSysColors/2; i++) {
        pe[i].peFlags = 0;
    }
    for (; i < iPalEntries-iSysColors/2; i++) {
        pe[i].peFlags = PC_NOCOLLAPSE;
    }
    for (; i < iPalEntries; i++) {
        pe[i].peFlags = 0;
    }

    // Resize the palette in case it was smaller
    m_pPalette->ResizePalette(iPalEntries);

    // Update the palette entries with what is now in the physical palette
    m_pPalette->SetPaletteEntries(0, iPalEntries, pe);

    ::ReleaseDC(hwndActive, hdcScreen);

	// Now we should take each frame and map its colors to the identity palette
	// but in fact since we basically grabbed the hardware palette anyway, the image
	// palette indices are not likely to change so we'll be bad (but very fast) and
	// not map the image pixels. See "Animation Techniques in Win32" for information
	// on how to map the pixel colors.


}

///////////////////////////////////////////////////////////////////////////////
// CMovie

CMovie::CMovie()
{
	m_bIsPlaying = FALSE;
	m_bIsRecording = FALSE;
	m_iCurFrame = 0;
}

BOOL CMovie::Create(C3dWnd* p3dWnd)
{
	ASSERT(p3dWnd);
	m_p3dWnd = p3dWnd;
	return TRUE;
}

BOOL CMovie::Play()
{
	Stop();

	// Start playback
	m_iCurFrame = 0;
	m_bIsPlaying = TRUE;
	return TRUE;
}

void CMovie::Stop()
{
	if (m_bIsPlaying) {
		m_bIsPlaying = FALSE;

		// Remap colors
		m_Take.Optimize();
	}
	if (m_bIsRecording) {
		m_bIsRecording = FALSE;
	}
}

BOOL CMovie::Record()
{
	Stop();

	// Delete any existing recording
	m_Take.DeleteAll();
	m_iCurFrame = 0;

	// Grab current palette
	ASSERT(m_p3dWnd);
	CDirectDraw* pDD = m_p3dWnd->GetDD();
	ASSERT(pDD);
	m_Take.SetPalette(pDD->GrabPalette());

	// Save frame size
	CRect rc;
	m_p3dWnd->GetClientRect(&rc);
	m_Take.SetSize(rc.Width(), rc.Height());

	// Start recording
	m_bIsRecording = TRUE;
	return TRUE;
}

BOOL CMovie::Update()
{
	if (m_bIsPlaying) {

		// Show next frame
		if (m_Take.IsEmpty()) {
			m_bIsPlaying = FALSE;
			return FALSE;
		}

		// Find current frame
		POSITION pos = m_Take.FindIndex(m_iCurFrame);
		if (!pos) {
			m_bIsPlaying = FALSE;
			return FALSE;
		}
		CMovieFrame* pFrame = m_Take.GetAt(pos);
		ASSERT(pFrame);

		// Get window device context (DC)
		ASSERT(m_p3dWnd);
		CDC* pdc = m_p3dWnd->GetDC();

		// Set up palette
		CPalette* pOldPal = NULL;
		CPalette* pPal = m_Take.GetPalette();
		if (pPal) {
			pOldPal = pdc->SelectPalette(pPal, FALSE);
			pdc->RealizePalette();
		}

		// Create memory DC for bitmap
		CDC dcMem;
		dcMem.CreateCompatibleDC(pdc);

		// Select palette into memory DC
		CPalette* pOldMemPal = NULL;
		if (pPal) {
			pOldMemPal = dcMem.SelectPalette(pPal, FALSE);
			dcMem.RealizePalette();
		}

		// Select bitmap and place into memory DC
		HBITMAP hOldBmp = (HBITMAP) ::SelectObject(dcMem, pFrame->m_hBmp);

		// Copy bitmap to window
		pdc->BitBlt(0, 0,
				    m_Take.GetWidth(), m_Take.GetHeight(),
				    &dcMem,
				    0, 0,
				    SRCCOPY);

		// Tidy up
		if (pOldMemPal) {
			dcMem.SelectPalette(pOldMemPal, FALSE);
		}
		::SelectObject(dcMem, hOldBmp);
		if (pOldPal) {
			pdc->SelectPalette(pOldPal, FALSE);
		}
		m_p3dWnd->ReleaseDC(pdc);

		m_iCurFrame++;
		return TRUE; // More frames

	}

	if (m_bIsRecording) {

		// Grab the next frame
		// Get a pointer to DirectDraw object in 3D window
		ASSERT(m_p3dWnd);
		CDirectDraw* pDD = m_p3dWnd->GetDD();
		ASSERT(pDD);

		// Grab current image
		BITMAPINFO* pBMI = NULL;
		BYTE* pBits = NULL;
		HBITMAP hBmp = pDD->GrabImage(&pBMI, (void**)&pBits);
		ASSERT(hBmp);
		ASSERT(pBMI);
		ASSERT(pBits);

		// Create movie frame
		CMovieFrame* pFrame = new CMovieFrame(hBmp, pBMI, pBits);

		// Add this frame to the take
		m_Take.AddTail(pFrame);

		m_iCurFrame++;
		return TRUE; // More frames
	}

	return FALSE; // Nothing to do
}

BOOL CMovie::Save()
{
	if (GetNumFrames() <= 0) return FALSE;

	// Get file name
	ASSERT(m_p3dWnd);
	CFileDialog dlg(FALSE,
					"avi",
					NULL,
					OFN_OVERWRITEPROMPT,
					"AVI Files (*.avi)|*.avi||",
					m_p3dWnd);
	if (dlg.DoModal() != IDOK) return FALSE;

	// Open AVI file
	HRESULT hr;
	PAVIFILE pfile = NULL;
	hr = ::AVIFileOpen(&pfile,
					   dlg.GetFileName(),
					   OF_CREATE | OF_WRITE,
					   NULL);
	if (FAILED(hr)) return FALSE;

	// Create video stream in file
	PAVISTREAM pstream = NULL;
	AVISTREAMINFO si;
	memset(&si, 0, sizeof(si));
	si.fccType = streamtypeVIDEO;
	si.fccHandler =  mmioFOURCC('M','S','V','C');
	si.dwRate = 100; // Fps
	si.dwScale = 1;
	si.dwLength = 0;
	si.dwQuality = (DWORD) -1;
	si.rcFrame.top = 0;
	si.rcFrame.left = 0;
	si.rcFrame.bottom = m_Take.GetHeight();
	si.rcFrame.right = m_Take.GetWidth();
	strcpy(si.szName, "3dPlus Movie");

	hr = ::AVIFileCreateStream(pfile, &pstream, &si);
	ASSERT(SUCCEEDED(hr));

	// Set format
	CMovieFrame* pFrame = m_Take.GetHead();
	ASSERT(pFrame);
	int iSize = sizeof(BITMAPINFOHEADER)
			  + DIBColorEntries((BITMAPINFOHEADER*)(pFrame->m_pBMI))
				* sizeof(RGBQUAD);
	hr = ::AVIStreamSetFormat(pstream,
							  0,
							  pFrame->m_pBMI,
							  iSize);
	ASSERT(SUCCEEDED(hr));

	// Write out frames
	POSITION pos = m_Take.GetHeadPosition();
	int iSample = 0;
	while (pos) {

		CMovieFrame* pFrame = m_Take.GetNext(pos);
		BITMAPINFOHEADER* pBIH = (BITMAPINFOHEADER*) &(pFrame->m_pBMI->bmiHeader);
		int iBits = DIBStorageWidth(pBIH) * pBIH->biHeight;
		hr = ::AVIStreamWrite(pstream, 
							  iSample, // This sample
							  1, // Just the one, thank you
							  pFrame->m_pBits, // The data
							  iBits, // Size of the buffer
							  0,
							  NULL,
							  NULL);
		ASSERT(SUCCEEDED(hr));
		iSample++;
	}

	// Release stream
	::AVIStreamRelease(pstream);

	// Release AVI stream and file
	::AVIFileRelease(pfile);

	return TRUE;
}





